class Error(Exception):
    """Base class for other exceptions"""
    pass


class EmptyFileError(Error):
    message = "File doesn't have contents"
    pass


class FalseError(Error):
    message = "Error in API response"
    pass


class FileProcessError(Error):
    pass

class InvalidDatetimeInHl7Message(Error):
    def __init__(self,child):
        Error.__init__(self,("Datetime in not valid in child : "+child))
        self.message = "Datetime in not valid in child : "+child


class InvalidLoginToken(Error):
    message = "Invalid Login Token"
    pass

class LoginTokenExpired(Error):
    message = "Login Token Expired"
    pass

class LicenseHasExpired(Error):
    message = "License Has Expired..Please check domain name in headers.."
    pass

class DeadlockCondition(Error):
    message = "Deadlock found when trying to get lock; try restarting transaction"
    pass

class LockWaitTimeout(Error):
    message = "Lock wait timeout exceeded; try restarting transaction"
    pass

class LoginFailed(Error):
    def __init__(self,msg):
        Error.__init__(self,(msg))
        self.message=msg

class InvalidNumbers(Error):
    def __init__(self, num):
        Error.__init__(self,('","'.join(num))+" are not numbers")
        self.message = ",".join(num)+" are not numbers"

    def __str__(self):
        # return self.num+" not an integer value"
        return self.message

class InvalidNumber(Error):
    def __init__(self, num):
        Error.__init__(self, "'"+num+"' is not a number")
        # self.num=num
        self.message = "'"+num+"' is not a number"

    def __str__(self):
        # return self.num+" not an integer value"
        return self.message


class InvalidCalculation(Error):
    def __init__(self, num):
        Error.__init__(self, num+" invalid calculation")

    def __str__(self):
        return "invalid calculation"

class HL7ParserError(Error):
    def __init__(self):
        Error.__init__(self,"Error in HL7 message Parsing")

class EventHandlingError(Error):
    def __init__(self):
        Error.__init__(self,"Error in Event Handling")
class InvalidInput(Error):
    def __init__(self,message):
        Error.__init__(self, message)
        self.message=message
    def __str__(self):
        return self.message

class InvalidDateTime(Error):
    def __init__(self):
        Error.__init__(self,"invalid datetime")

    def __str__(self):
        return "invalid datetime"

class InvalidJob(Error):
    """
    Job body is not valid json or it doesn't have required fields
    """

    def __init__(self, jib, missing_parameters=None):
        self.message = "Invalid Job : [JID :(" + jib + ")" + ", Missing Parameters : (" + "".join(
            missing_parameters) + ")" + "]"


class MaxApiRetriesExceeded(Error):
    def __init__(self,message_details=None):
        self.message = "Max retries for a URL exceeded.There might be any exception. Please check."
        self.message_details = message_details

class BeanstalkConnectionFailed(Error):
    def __init__(self,message_details):
        self.message="Beanstalk Connection Failed"
        self.message_details=message_details


class RsysLogConnectionFailed(Error):
    def __init__(self,message_details):
        self.message="Rsyslog Connection Failed"
        self.message_details=message_details


class RedisConnectionFailed(Error):
    def __init__(self,message_details):
        self.message="Redis Connection Failed"
        self.message_details=message_details


class ResourceNotFound(Error):
    def __init__(self,message_details):
        self.message="Resource Not Found"
        self.message_details=message_details


class InternalServerError(Error):
    def __init__(self,message_details):
        self.message="Internal Server Error"
        self.message_details=message_details

class InvalidRequest(Error):
    def __init__(self,message_details):
        self.message="Invalid Request Sent"
        self.message_details=message_details


import os
from dotenv import load_dotenv

class EnvVariables(object):
    if(os.path.exists("../../.env")):
        load_dotenv()

    @staticmethod
    def get_val(key):
        return os.environ[key] if key in os.environ else None

    @staticmethod
    def get_localhost_url():
        return EnvVariables.get_val('LOCALHOST_URL')

    @staticmethod
    def get_api_base_url():
        return EnvVariables.get_val('BASE_URL')

    @staticmethod
    def get_api_version():
        return EnvVariables.get_val('API_VERSION')

    @staticmethod
    def get_api_17_version():
        return EnvVariables.get_val('API_VERSION_17')

    @staticmethod
    def get_admin_email():
        return EnvVariables.get_val('ADMIN_EMAIL')

    @staticmethod
    def get_admin_password():
        return EnvVariables.get_val('ADMIN_PASSWORD')

    @staticmethod
    def get_business_tag_id():
        return EnvVariables.get_val('BUSINESS_TAG_ID')

    @staticmethod
    def get_business_domain():
        return EnvVariables.get_val('BUSINESS_DOMAIN')

    @staticmethod
    def get_redis_host():
        return EnvVariables.get_val('REDIS_HOST')

    @staticmethod
    def get_redis_port():
        return EnvVariables.get_val('REDIS_PORT')

    @staticmethod
    def get_redis_db():
        return EnvVariables.get_val('REDIS_DB')

    @staticmethod
    def get_beanstalk_host():
        return EnvVariables.get_val('BEANSTALK_HOST')

    @staticmethod
    def get_beanstalk_port():
        return EnvVariables.get_val('BEANSTALK_PORT')

    @staticmethod
    def get_syslog_hanlder_address():
        return EnvVariables.get_val('SYSLOG_HANDLER_ADDRESS')

    @staticmethod
    def get_syslog_hanlder_port():
        return EnvVariables.get_val('SYSLOG_HANDLER_PORT')

    @staticmethod
    def get_csvfilepath():
        return EnvVariables.get_val('CSVFILEPATH')

    @staticmethod
    def get_workers():
        return EnvVariables.get_val('PARALLEL_PROCESS_WORKER_COUNT')


"""
    Logging module
"""
import logging
import logging.handlers
import socket
import json
import sys
import os
import traceback
import datetime
from ZvolvArithmetic.zauthentication.server_configuration import EnvVariables
from ZvolvArithmetic.zauthentication import user_defined_exceptions

def format_exception(e):
    """
    This function will format the exception message into a list and return it
    :param e: Exception
    :return:
    """
    exception_list = traceback.format_stack()
    exception_list = exception_list[:-2]
    exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
    exception_list.extend(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1]))
    # print exception_list
    return exception_list


class LoggerClass:
    """
    This class will create the singleton logger object with logger dictionary which will map the logger name with logging object

    Exmaple:
        -API Logging:
            api_log({"message":"API Logs","url":BASE_URL,"body":body,"method":method})

        -Queue Logging:
            queue_log({"message": "Queue Logs ","queue": "tube_WF", "zviceId":body['BusinessTag']})

        -python logging:
            pythonLog({"message": "Python Logs,"zviceId":body['BusinessTag']})

        -Business Specific Logging:
            log(body['BusinessTag'],{"message":"Business Logs","zviceID":body['BusinessTag']})
        This function has two parameters as it will create business.log file at runtime

    """
    __instance = None  # private variable

    def __init__(self):

        if LoggerClass.__instance:  # If instance is already present
            raise Exception("This class is a singleton!")
        else:
            LoggerClass.__instance = self
            LoggerClass.__instance.loggers = {}  # dictionary of logger_name and logging object as key and value


    @staticmethod
    def create_logger(app_name, log_level=logging.DEBUG, stdout=False, syslog=True, file=False):
        """
        create logging object with logging to syslog, file and stdout
        :param app_name app name
        :param log_level logging log level
        :param stdout log to stdout
        :param syslog log to syslog
        :param file log to file
        :return: logging object
        """
        # disable requests logging
        # logging.getLogger("requests").setLevel(logging.ERROR)
        # logging.getLogger("urllib3").setLevel(logging.ERROR)

        # create logger
        logger = logging.getLogger(app_name)
        logger.setLevel(log_level)

        # set log format to handlers
        formatter = logging.Formatter('%(name)s - %(asctime)s - %(levelname)s - %(message)s')
        path = "/var/log/script-logs/script-logs.log"
        if file:
            # create file logger handler
            fh = logging.FileHandler('my-sample-app.log')
            fh.setLevel(log_level)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

        if syslog:
            # create syslog logger handler
            env_syslog_hanlder_address, env_syslog_hanlder_port = EnvVariables.get_syslog_hanlder_address(), EnvVariables.get_syslog_hanlder_port()
            try:
                syslog_hanlder_address = (env_syslog_hanlder_address, int(
                    env_syslog_hanlder_port)) if env_syslog_hanlder_port else env_syslog_hanlder_address

            except (ValueError,TypeError):
                raise user_defined_exceptions.RsysLogConnectionFailed(message_details=
                                                                {"SYSLOG_HANDLER_ADDRESS":EnvVariables.get_syslog_hanlder_address(),
                                                                 "SYSLOG_HANDLER_PORT":EnvVariables.get_syslog_hanlder_port()})
            try:
                sh = logging.handlers.SysLogHandler(address=syslog_hanlder_address)
            except socket.error:
                raise user_defined_exceptions.RsysLogConnectionFailed(message_details={"SYSLOG_HANDLER_ADDRESS":env_syslog_hanlder_address})
            sh.setLevel(log_level)
            sf = logging.Formatter('%(name)s: %(message)s')
            sh.setFormatter(sf)
            logger.addHandler(sh)

        if stdout:
            # create stream logger handler
            ch = logging.StreamHandler()
            ch.setLevel(log_level)
            ch.setFormatter(formatter)
            logger.addHandler(ch)

        return logger

    @staticmethod
    def error(logger_name,log_message_dict):
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "ERROR")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))
        log.error(message)

    @staticmethod
    def debug(logger_name,log_message_dict):
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "DEBUG")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))
        log.debug(message)

    @staticmethod
    def warn(logger_name,log_message_dict):
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "WARN")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))
        log.warn(message)

    @staticmethod
    def warning(logger_name,log_message_dict):
        """
        This function will print the str object of dict(logging message's dict) in given loggername's file
        :param logger_name: api_log,pythonLog,etc.,
        :param dict:
        :return:
        """
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "WARNING")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))

        log.warning(message)

    @staticmethod
    def info(logger_name,log_message_dict):
        """
        This function will print the str object of dict(logging message's dict) in given loggername's file
        :param logger_name: api_log,pythonLog,etc.,
        :param dict:
        :return:
        """
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "INFO")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))
        log.info(message)

    @staticmethod
    def exception_log(logger_name,log_message_dict):
        e = log_message_dict['exception']
        log_message_dict['exception'] = {  # formatting exception object to avoid json serializable error
            "error": e.__class__.__name__,
            "args": str(e.args),
            "message": str(e.message),
            "traceback": format_exception(e)
        }
        log = LoggerClass.get_log(logger_name)
        log_message_dict = update_log_message_dict(log_message_dict, "ERROR")
        message = (json.dumps(log_message_dict))[:127999] if len(json.dumps(log_message_dict)) > 128000 else (json.dumps(log_message_dict))
        log.error(message)

    @classmethod
    def get_log(cls, logger_name):
        """
        This function will search the logger name in logger dictionaries keys.
        If it is present then it will return the corresponding logger object
        else it will create object using setup_logger()
        :param logger_name: String
            Example : "root","api_log"
        :return:
        """
        if not cls.__instance:
            LoggerClass()
        if logger_name in cls.__instance.loggers.keys():  # the log object already exists in logger dict
            return cls.__instance.loggers[logger_name]
        logger = LoggerClass.create_logger(logger_name)
        cls.__instance.loggers[logger_name] = logger
        return logger


def update_log_message_dict(log_message_dict, log_level):
    """
    Returns updated log_message_dict
    :param log_message_dict: message dictionary
    :param log_level: logging level
    :return:
    """
    new_contents = {'log_level': log_level, 'DateCreated': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),"pid":os.getpid()}
    log_message_dict.update(new_contents)
    return log_message_dict


import copy
import os
from ZvolvArithmetic.zauthentication.zsession import generate_request_headers

class Session:
    """
    A session stores configuration state and allows you to create service

    :type zvolv_access_key_id: string
    :param zvolv_access_key_id: Email access key ID
    :type zvolv_secret_access_key: string
    :param zvolv_secret_access_key: password secret access key
    :type zvolv_domain_context: string
    :param zvolv_domain_context: zvolv business domain
    """

    def __init__(
        self,
        zvolv_access_key_id=None,
        zvolv_secret_access_key=None,
        zvolv_domain_context =None,
    ):
        if zvolv_access_key_id or zvolv_secret_access_key or zvolv_domain_context:
            self.set_credentials(
               zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context,
            )
    
    def set_credentials(self,zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context):
        print(generate_request_headers(zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context))


import logging
from ZvolvArithmetic.zauthentication.session1 import Session

__author__ = 'Zvolv'
__version__ = '0.0.1'

# The default zvolv session; autoloaded when needed.
DEFAULT_SESSION = None

def setup_default_session(**kwargs):
    """
    Set up a default session, passing through any parameters to the session
    constructor. There is no need to call this unless you wish to pass custom
    parameters, because a default session will be created for you.
    """
    global DEFAULT_SESSION
    DEFAULT_SESSION = Session(**kwargs)

def _get_default_session():
    """
    Get the default session, creating one if needed.

    :rtype: :py:class:`~.`
    :return: The default session
    """
    if DEFAULT_SESSION is None:
        setup_default_session()
    return DEFAULT_SESSION

setup_default_session(zvolv_access_key_id='yogesh.rjadhav22@gmail.com',
        zvolv_secret_access_key='Pucsd!@3',
        zvolv_domain_context ='yogeshjadhav')


import json
import os
import hashlib
import time
import sys
from ZvolvArithmetic.zauthentication.server_configuration import EnvVariables
from ZvolvArithmetic.zauthentication import zexcute_apis
from ZvolvArithmetic.zauthentication.loggerClass import LoggerClass as LogClass

folder_path = "/tmp"            # For server
# folder_path = os.getcwd()     # For local testing
CACHE_FILE_NAME = 'token_cache.json'
CACHE_FILE_PATH = os.path.join(folder_path, CACHE_FILE_NAME)
zvolv_business_tag_id = "98NCMBD2KBZ4R"

def load_token_cache():
    # Checking whether file is present or not
    if not os.path.isfile(CACHE_FILE_PATH):
        # Creating file with value {}
        with open(CACHE_FILE_PATH, 'w') as file:
            json.dump({}, file)

    # Reading value from file
    with open(CACHE_FILE_PATH, 'r') as cache_file:
        return json.load(cache_file)


def save_token_cache(token_cache):
    with open(CACHE_FILE_PATH, 'w') as cache_file:
        json.dump(token_cache, cache_file)


def zlogin(zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context, zvolv_business_tag_id):
    token_cache = load_token_cache()
    business_domain, business_tag = zvolv_domain_context, zvolv_business_tag_id

    # Check if the token is already created and stored. Also, check if it should not be older than 60 seconds.
    if token_cache and business_domain in token_cache and token_cache and (time.time() - token_cache[business_domain]["serverTime"]) < 60:
        return token_cache[business_domain]

    # sha512pwd = hashlib.sha512(EnvVariables.get_admin_password()).hexdigest()
    # payload = {'email': EnvVariables.get_admin_email(), 'password': sha512pwd}

    email_id = zvolv_access_key_id
    sha512pwd = hashlib.sha512(zvolv_secret_access_key.encode('utf-8')).hexdigest()
    payload = {'email': email_id, 'password': sha512pwd}
    headers = {'businessDomain': business_domain, 'jwt': 'true', 'businessTagID': business_tag ,'Device':'browser'}
    # login_url = EnvVariables.get_localhost_url() + EnvVariables.get_api_17_version() + "user/login"
    login_url = "https://zvolv.co/rest/v17/user/login"

    login_response = zexcute_apis.execute(rest_url=login_url, method="POST", data=payload, headers=headers)
    # LogClass.warning("api-logs", {"message": "Logging Response", "response": login_response})

    if login_response:
        # Create login data to store in file
        token_cache = {
            business_domain: {}
        }
        for key, value in login_response.items():
            token_cache[business_domain][key] = value
        save_token_cache(token_cache)
        return login_response
    else:
        raise Exception('Failed to get login token')


def get_request_configs():
    if request:
        if request.headers.get("Businessdomain") == "lk-nso":
            return request.headers.get("Businessdomain"), "W4S3L2ZEFRVT2"
        else:
            return request.headers.get("Businessdomain"), "98NCMBD2KBZ4R"
    else:
        return "yogeshjadhav","98NCMBD2KBZ4R"

def generate_login_token(zvolv_access_key_id,zvolv_secret_access_key, zvolv_domain_context, zvolv_business_tag_id):
    login = zlogin(zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context, zvolv_business_tag_id)
    print(login)
    return login["loginToken"]


def generate_request_headers(zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context):
    business_domain = zvolv_domain_context
    auth_key = generate_login_token(zvolv_access_key_id, zvolv_secret_access_key, zvolv_domain_context, zvolv_business_tag_id)
    headers = {'domain': business_domain, 'Authorization': 'bearer ' + auth_key, 'Content-Type': 'application/json'}
    return headers


import json
import urllib
import requests
from pprint import pprint
from ZvolvArithmetic.zauthentication.loggerClass import LoggerClass as LogClass
from ZvolvArithmetic.zauthentication import user_defined_exceptions


def handle_request(func):
    def wrapper_func(*args, **kwargs):
        # LogClass.warning("script-logs", {"message": "api request", "request_details": kwargs})
        query_params = kwargs.get("query_params")
        if query_params:
            updated_query_params = [(k, str(v).lower() if isinstance(v, bool) else v) for k, v in query_params.items()]
            encoded_query_params = urllib.urlencode(updated_query_params)
            kwargs["rest_url"] += "?" + encoded_query_params
        headers = kwargs.get("headers")
        if headers is None:
            import zsession
            kwargs["headers"] = zsession.generate_request_headers()

        if kwargs.get("json"):
            kwargs["headers"].update({'Content-Type': 'application/json'})
            kwargs["data"] = json.dumps(kwargs["json"])
        ret_val = func(*args, **kwargs)
        return ret_val

    return wrapper_func


def handle_response(func):
    def wrapper_func(*args, **kwargs):
        ret_val = func(*args, **kwargs)
        status_code, response = ret_val.status_code, ret_val.text
        response_details = {"status_code": status_code, "response": response}
        response_details.update(kwargs)
        # LogClass.warning("script-logs", {"message": "api response", "response_details": response_details})
        if status_code == 404:
            raise user_defined_exceptions.ResourceNotFound(message_details={"URL": kwargs.get("rest_url")})
        elif status_code == 500:
            raise user_defined_exceptions.InternalServerError(message_details={"URL": kwargs.get("rest_url")})
        elif status_code == 200:
            if json.loads(response).get("error"):
                raise user_defined_exceptions.InvalidRequest(
                    message_details={"error_message": json.loads(response).get("message")})
            return json.loads(response)
        else:
            return json.loads(response)

    return wrapper_func


@handle_request
@handle_response
def execute(rest_url, method, query_params=None, data=None, json=None, headers=None):
    print("checked execute")
    """
    non-used params have manipulated in decorators.
    :param rest_url:
    :param method:
    :param query_params:
    :param data:
    :param json:
    :param headers:
    :return:
    """
    if method == "GET":
        return requests.get(url=rest_url, headers=headers)
    elif method == "POST":
        return requests.post(url=rest_url, data=data, headers=headers)
    elif method == "PUT":
        return requests.put(url=rest_url, data=data, headers=headers)